import com.fasterxml.uuid.Generators;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.LongBuffer;
import java.util.StringTokenizer;
import java.util.UUID;

public class UUIDUtil {
    private static final int[] uuidLength = {8, 4, 4, 4, 12};


    public static UUID generateUUIDByName(String name) {
        return Generators.nameBasedGenerator().generate(name);
    }


    public static UUID generateUUIDRandomly() {
        return Generators.randomBasedGenerator().generate();
    }

    public static UUID generateUUIDTimely() {

        UUID uuid = Generators.timeBasedGenerator().generate();
        final ByteBuffer sourceBuffer = ByteBuffer.allocateDirect(16);

        int time_low = (int) (uuid.getMostSignificantBits() >> 32);
        short time_mid = (short) ((uuid.getMostSignificantBits() >> 16) & 0xffff);
        short time_high_and_version = (short) ((uuid.getMostSignificantBits()) & 0xffff);
        short variant_and_sequence = (short) ((uuid.getLeastSignificantBits() >> 48) & 0xffff);
        short node_most = (short) ((uuid.getLeastSignificantBits() >> 32) & 0xffff);
        int node_least = (int) uuid.getLeastSignificantBits();

        sourceBuffer.clear();
        sourceBuffer.putShort(time_high_and_version);
        sourceBuffer.putShort(time_mid);
        sourceBuffer.putInt(time_low);
        sourceBuffer.putShort(variant_and_sequence);
        sourceBuffer.putShort(node_most);
        sourceBuffer.putInt(node_least);
        sourceBuffer.flip();

//            return time_high_and_version + '-' + time_mid + '-' + time_low + '-' + variant_and_sequence + '-' + node;

        return new UUID(sourceBuffer.getLong(), sourceBuffer.getLong());

    }

    public static String generateUUIDSerialTimely() {
        final UUID uuid = generateUUIDTimely();
        return UUIDToString(uuid);
    }


    public static String UUIDToBase64(final UUID orgUUID) {
        if (orgUUID == null) {
            return null;
        }
        final long most = orgUUID.getMostSignificantBits();
        final long last = orgUUID.getLeastSignificantBits();
        return new String(Base64.encode(ByteUtil.toBytes(most, last)));
    }


    public static UUID base64ToUUID(final String uuidString) {

        if (uuidString == null) {
            return null;
        }
        try {
            byte[] decoded = Base64.decode(uuidString);
            if (decoded == null) {
                return null;
            }
            final ByteBuffer sourceBuffer = ByteBuffer.allocateDirect(16);
            sourceBuffer.clear();
            sourceBuffer.put(decoded);
            sourceBuffer.flip();
            return new UUID(sourceBuffer.getLong(), sourceBuffer.getLong());
        } catch (Exception iae) {
            return null;
        }
    }


    public static String generateUUIDBase58() {
        final UUID uuid = generateUUIDTimely();
        return UUIDToBase58(uuid);
    }

    public static String generateUUIDBase58Checked() {
        final UUID uuid = generateUUIDTimely();
        return UUIDToBase58Checked(uuid);
    }

    public static byte[] UUIDToBytes(final UUID orgUUID) {
        if (orgUUID == null) {
            return null;
        }
        final long most = orgUUID.getMostSignificantBits();
        final long last = orgUUID.getLeastSignificantBits();
        return ByteUtil.toBytes(most, last);
    }


    public static String UUIDToBase58(final UUID orgUUID) {
        if (orgUUID == null) {
            return null;
        }
        final long most = orgUUID.getMostSignificantBits();
        final long last = orgUUID.getLeastSignificantBits();
        return Base58.encode(ByteUtil.toBytes(most, last));
    }


    public static String UUIDToBase58Checked(final UUID orgUUID) {
        if (orgUUID == null) {
            return null;
        }
        final long most = orgUUID.getMostSignificantBits();
        final long last = orgUUID.getLeastSignificantBits();
        return Base58.encodeChecked(ByteUtil.toBytes(most, last));
    }

    public static UUID base58ToUUID(final String uuidString) {

        if (uuidString == null) {
            return null;
        }
        try {

            final ByteBuffer sourceBuffer = ByteBuffer.allocateDirect(16);
            sourceBuffer.clear();
            sourceBuffer.put(Base58.decode(uuidString));
            sourceBuffer.flip();
            return new UUID(sourceBuffer.getLong(), sourceBuffer.getLong());
        } catch (IllegalArgumentException iae) {
            return null;
        }
    }

    public static UUID base58ToUUIDChecked(final String uuidString) {

        if (uuidString == null) {
            return null;
        }
        byte[] decoded = Base58.decodeChecked(uuidString);
        if (decoded == null) {
            return null;
        }
        final ByteBuffer sourceBuffer = ByteBuffer.allocateDirect(16);
        sourceBuffer.clear();
        sourceBuffer.put(decoded);
        sourceBuffer.flip();
        return new UUID(sourceBuffer.getLong(), sourceBuffer.getLong());
    }

    public static UUID bytesToUUID(byte[] uuidByte) {
        if (uuidByte == null || uuidByte.length != 16) {
            return null;
        }
        final long most = ByteUtil.fromBytesAsLongMost(uuidByte);
        final long last = ByteUtil.fromBytesAsLongLast(uuidByte);
        return new UUID(most, last);
    }


    public static String UUIDToString(final UUID orgUUID) {

        if (orgUUID == null) {
            return null;
        }
        final CharBuffer sourceBuffer = ByteBuffer.allocateDirect(Character.BYTES * 32).asCharBuffer();
        sourceBuffer.clear();

        final String orgStr = orgUUID.toString();
        final StringTokenizer orgStrings = new StringTokenizer(orgStr, "-");
        while (orgStrings.hasMoreElements()) {
            sourceBuffer.put(orgStrings.nextToken());
        }
        sourceBuffer.flip();
        return sourceBuffer.toString();
    }

    public static UUID stringToUUID(final String uuidString) {
        if (uuidString == null) {
            return null;
        }
        switch (uuidString.length()) {
            case 36:
                return UUID.fromString(uuidString);
            case 32:
                final CharBuffer sourceBuffer = ByteBuffer.allocateDirect(Character.BYTES * uuidString.length()).asCharBuffer();
                final LongBuffer targetBuffer = ByteBuffer.allocateDirect(Long.BYTES * uuidLength.length).asLongBuffer();
                sourceBuffer.clear();
                sourceBuffer.put(uuidString);
                sourceBuffer.flip();
                targetBuffer.clear();

                for (int size : uuidLength) {
                    final StringBuilder builder = new StringBuilder();
                    final char[] buf = new char[size];
                    builder.append("0x");
                    sourceBuffer.get(buf, 0, size);
                    builder.append(buf);
                    targetBuffer.put(Long.decode(builder.toString()));
                }
                targetBuffer.flip();
                long mostSigBits = targetBuffer.get();
                mostSigBits <<= 16;

                mostSigBits |= targetBuffer.get();
                mostSigBits <<= 16;

                mostSigBits |= targetBuffer.get();

                long leastSigBits = targetBuffer.get();
                leastSigBits <<= 48;
                leastSigBits |= targetBuffer.get();

                return new UUID(mostSigBits, leastSigBits);
            default:
                return null;
        }
    }
}

